from __future__ import annotations

from .._internal import _cached_toml, _writers
from .config import WriterConfig, WriteVersionAssignSettings, WriteVersionTomlSettings


class Writer:
    """
    A Writer is an interface for data writing capabilities.

    Typically, users should simply create a DJ and then access the writer built by it,
    as in:

    .. code-block:: pycon

        >>> from mddj.api import DJ
        >>> dj = DJ()
        >>> dj.write.version("0.2.0")
        '0.1.0'
    """

    def __init__(
        self,
        config: WriterConfig,
        document_cache: _cached_toml.TomlDocumentCache | None = None,
    ) -> None:
        self.config = config
        self._document_cache = document_cache or _cached_toml.TomlDocumentCache()

    def version(self, new_version: str) -> str:
        """
        Write a new version into the project metadata, returning the old version.

        If no previous version can be found, a ``LookupError`` is raised, on the
        presumption that this means that the writer is not correctly configured for
        the project.
        """
        write_version_settings = self.config.write_version_settings
        file_path = self.config.project_directory / write_version_settings.file_path

        if isinstance(write_version_settings, WriteVersionAssignSettings):
            result = _writers.write_simple_assignment(
                file_path, write_version_settings.key, new_version
            )
        elif isinstance(write_version_settings, WriteVersionTomlSettings):
            return _writers.write_toml_value(
                file_path,
                write_version_settings.toml_path,
                new_version,
                loaded_document=self._document_cache.load(file_path),
            )
        else:
            raise NotImplementedError(
                "Unrecognized write_version_config. "
                f"write_version={self.config.write_version}"
            )

        if result is None:
            raise LookupError(
                "No previous version was found, so no update was written."
            )

        return result



import typing as t

import click
from packaging.version import Version

from mddj._cli.state import CommandState, common_args


@click.command("version")
@common_args
@click.option(
    "--attr",
    help=(
        "Read an attribute of the version, not the whole version number. "
        "If 'pre' or 'post' is used and the release is not a prerelease or "
        "post-release, the command will exit(1) with an error."
    ),
    type=click.Choice(
        ("release", "major", "minor", "micro", "pre", "post"), case_sensitive=False
    ),
)
def read_version(
    *,
    attr: t.Literal["release", "major", "minor", "micro", "pre", "post"] | None,
    state: CommandState,
) -> None:
    """Read the 'Version' of the current project."""
    version = state.dj.read.version()
    if attr is not None:
        version = _get_version_attr(version, attr)
    click.echo(version)


def _get_version_attr(
    version: str, attr: t.Literal["release", "major", "minor", "micro", "pre", "post"]
) -> str:
    parsed_version = Version(version)

    if attr == "release":
        return ".".join(str(n) for n in parsed_version.release)
    elif attr == "major":
        return str(parsed_version.major)
    elif attr == "minor":
        return str(parsed_version.minor)
    elif attr == "micro":
        return str(parsed_version.micro)
    elif attr == "pre":
        if not parsed_version.is_prerelease:
            click.echo(f"'{version}' is not a prerelease", err=True)
            click.get_current_context().exit(1)
        pre_parts: tuple[str | int, ...] = (
            parsed_version.pre  # type: ignore[assignment]
        )
        return "".join(str(n) for n in pre_parts)
    elif attr == "post":
        if not parsed_version.is_postrelease:
            click.echo(f"'{version}' is not a post-release", err=True)
            click.get_current_context().exit(1)
        return str(parsed_version.post)
    else:
        t.assert_never(attr)



import click

from mddj._cli.state import CommandState, common_args

from .list_versions import tox_list_versions
from .min_version import tox_min_version


@click.group("tox")
@common_args
def read_tox(*, state: CommandState) -> None:
    """Read metadata from the current project via 'tox'."""


read_tox.add_command(tox_min_version)
read_tox.add_command(tox_list_versions)



import click

from mddj._cli.state import CommandState, common_args


@click.command("requires-python")
@click.option(
    "--lower-bound",
    is_flag=True,
    help="Show a lower bound rather than the full Requires-Python data.",
)
@common_args
def read_requires_python(*, state: CommandState, lower_bound: bool) -> None:
    """
    Read the 'Requires-Python' data.
    """
    requires_python = state.dj.read.requires_python(lower_bound=lower_bound)
    click.echo(requires_python)



from __future__ import annotations

import pathlib

import tomlkit


class TomlDocumentCache:
    def __init__(self) -> None:
        self._cache: dict[pathlib.Path, tomlkit.TOMLDocument] = {}

    def load(self, path: pathlib.Path) -> tomlkit.TOMLDocument:
        if not path.is_absolute():
            raise ValueError("Cached loads must use absolute paths for consistency.")

        if path not in self._cache:
            with path.open("r", encoding="utf-8") as fp:
                data = tomlkit.load(fp)
            self._cache[path] = data
        return self._cache[path]



import contextlib
import os
import sys

import pytest


# remove this fixture once Python 3.10 support is done
@pytest.fixture(scope="session")
def chdir():
    if sys.version_info >= (3, 11):
        return contextlib.chdir

    # based on the contents of contextlib for py3.11+
    class chdir:
        def __init__(self, path) -> None:
            self.path = path
            self._old_cwd = []

        def __enter__(self):
            self._old_cwd.append(os.getcwd())
            os.chdir(self.path)

        def __exit__(self, *excinfo):
            os.chdir(self._old_cwd.pop())

    return chdir



from __future__ import annotations

import sys
import typing as t

import tomlkit

if sys.version_info >= (3, 13):
    from typing import TypeIs
else:
    from typing_extensions import TypeIs

TomlValue: t.TypeAlias = tomlkit.items.Item | tomlkit.container.Container
TomlTable: t.TypeAlias = tomlkit.items.Table | tomlkit.items.InlineTable
TomlArray: t.TypeAlias = tomlkit.items.Array | tomlkit.items.AoT
TomlMapping: t.TypeAlias = tomlkit.TOMLDocument | TomlTable


def is_toml_array(obj: t.Any) -> TypeIs[TomlArray]:
    return isinstance(obj, (tomlkit.items.Array, tomlkit.items.AoT))


def is_toml_table(obj: t.Any) -> TypeIs[TomlTable]:
    return isinstance(obj, (tomlkit.items.Table, tomlkit.items.InlineTable))


def is_toml_mapping(obj: t.Any) -> TypeIs[TomlMapping]:
    return isinstance(
        obj, (tomlkit.TOMLDocument, tomlkit.items.Table, tomlkit.items.InlineTable)
    )



from __future__ import annotations

import functools
import re
import shutil
import subprocess
import sys


class ToxReaderError(RuntimeError):
    """The class of errors which can be raised if ``tox`` data discovery fails."""


class ToxReader:
    """
    A ToxReader is a specialized data reader for ``tox`` data.

    Typically, users should simply create a DJ and then access the tox reader built by
    it, as in:

    .. code-block:: pycon

        >>> from mddj.api import DJ
        >>> dj = DJ()
        >>> dj.read.tox.list_versions()
        ['3.8', '3.9', '3.10', '3.11', '3.12', '3.13']
    """

    @functools.cached_property
    def _tox_cmd(self) -> tuple[str, ...]:
        try:
            import tox  # noqa: F401
        except ImportError:
            tox_cmd = shutil.which("tox")
            if not tox_cmd:
                raise ToxReaderError(
                    "Cannot fetch tox data. A 'tox' command was not found."
                )
            return (tox_cmd,)
        else:
            return (sys.executable, "-m", "tox")

    @functools.cached_property
    def _tox_version(self) -> str:
        tox_version_proc = subprocess.run(
            [*self._tox_cmd, "--version"], text=True, capture_output=True
        )
        if tox_version_proc.returncode != 0:
            raise ToxReaderError("Cannot fetch tox data. 'tox --version' failed.")

        return tox_version_proc.stdout.strip().partition(" ")[0]

    def _check_tox_version(self) -> str:
        """
        Check that the 'tox' command is a supported version and return it.

        Any unexpected outputs will raise an error, which allows for a cleaner early
        abort.
        """
        from packaging.version import Version

        version = Version(self._tox_version)
        if version.major in (3, 4):
            return self._tox_version
        else:
            raise ToxReaderError("'tox --version' was not a recognized version.")

    @functools.cached_property
    def _tox_listenvs(self) -> tuple[str, ...]:
        self._check_tox_version()
        output = subprocess.check_output([*self._tox_cmd, "--listenvs"], text=True)
        return tuple(line for line in output.splitlines() if line)

    def list_python_versions(self) -> list[str]:
        """List the Python versions which appear in the list of tox environments."""
        versions = set()
        for envname in self._tox_listenvs:
            for part in envname.split("-"):
                if match := re.match(r"py(\d)\.?(\d+)", part):
                    versions.add(match.group(1) + "." + match.group(2))
        return list(versions)



import pytest

from mddj._internal._toml_path import ParseError, parse_toml_path


@pytest.mark.parametrize(
    "path, expect_result",
    [
        ("foo", ["foo"]),
        ("foo.bar", ["foo", "bar"]),
        ("foo.0.bar", ["foo", 0, "bar"]),
        ("foo.'0'.bar", ["foo", "0", "bar"]),
        ("foo\\.bar", ["foo.bar"]),
        ("foo\\\\.bar", ["foo\\", "bar"]),
        ("", []),
        ("''", [""]),
    ],
)
def test_valid_parse(path, expect_result):
    assert parse_toml_path(path) == expect_result


@pytest.mark.parametrize(
    "path, expect_message",
    [
        ("foo\\", "Terminal escape\\."),
        ("abc.", "Terminal '\\.' separator\\."),
        ("x..y", "Double '\\.' separator\\."),
        (
            "xy.foo'bar'",
            (
                "Quote chars within keys must be escaped "
                "or the entire key must be quoted\\."
            ),
        ),
        ("foo.'bar", "Unclosed quoted string\\."),
        ('foo."bar', "Unclosed quoted string\\."),
        ('foo."bar', "Unclosed quoted string\\."),
        (".x", "Leading '\\.' separator\\."),
    ],
)
def test_parse_error(path, expect_message):
    with pytest.raises(ParseError, match=expect_message):
        parse_toml_path(path)



import click

from mddj._cli.state import CommandState, common_args

from .dependencies import read_dependencies
from .requires_python import read_requires_python
from .tox import read_tox
from .version import read_version


@click.group("read")
@common_args
def read(*, state: CommandState) -> None:
    """Read metadata from the current project."""


read.add_command(read_requires_python)
read.add_command(read_version)
read.add_command(read_dependencies)
read.add_command(read_tox)



import typing as t


class ParseError(ValueError):
    pass


def parse_toml_path(path: str) -> list[str | int]:
    return list(_tokenize(path))


def _tokenize(path: str) -> t.Iterator[str | int]:
    for token, is_quoted in _scan(path):
        if is_quoted:
            yield token
        else:
            try:
                yield int(token)
            except ValueError:
                yield token


def _scan(path: str) -> t.Iterator[tuple[str, bool]]:
    if not path:
        return
    if path[0] == ".":
        raise ParseError("Leading '.' separator.")

    charstack = []

    quote: str | None = None
    # None = default state
    # s = skip next
    # e = reading an escaped char
    state: t.Literal["s", "e"] | None = None
    for i, char in enumerate(path):
        if state == "s":
            state = None
            continue

        try:
            peek: str | None = path[i + 1]
        except IndexError:
            peek = None

        if state == "e":
            charstack.append(char)
            state = None
            continue

        if char == "\\":
            state = "e"
            if peek is None:
                raise ParseError("Terminal escape.")
            continue

        if quote:
            if char != quote:
                charstack.append(char)
            else:
                if peek not in (".", None):
                    raise ParseError("Closed quote was not at end of key.")
                yield "".join(charstack), True
                charstack.clear()

                quote = None
                state = None
            continue

        if char == ".":
            if peek is None:
                raise ParseError("Terminal '.' separator.")
            if peek == ".":
                raise ParseError("Double '.' separator.")
            # skip if we haven't accumulated any chars, which can only happen
            # if we just finished a quoted string
            if not charstack:
                continue
            yield "".join(charstack), False
            charstack.clear()
        elif char in ("'", '"'):
            if charstack:
                raise ParseError(
                    "Quote chars within keys must be escaped "
                    "or the entire key must be quoted."
                )
            quote = char
        else:
            charstack.append(char)

    if quote:
        raise ParseError("Unclosed quoted string.")

    if charstack:
        yield "".join(charstack), False



import click
from packaging.version import Version

from mddj._cli.state import CommandState, common_args


@click.command("min-version")
@common_args
def tox_min_version(*, state: CommandState) -> None:
    """Print the minimum version of python tested under tox."""
    versions = state.dj.read.tox.list_python_versions()

    if not versions:
        raise RuntimeError("No tox version data found.")
    click.echo(str(sorted(Version(v) for v in versions)[0]))



import pytest

from mddj._internal._discovery import discover_project_dir


@pytest.mark.parametrize(
    "indicator_file", ("pyproject.toml", "setup.cfg", "setup.py", "tox.ini")
)
def test_can_discover_unambiguous_project_dir(tmp_path, indicator_file):
    (tmp_path / indicator_file).touch()

    working_dir = tmp_path / "subdir"
    working_dir.mkdir()
    assert str(discover_project_dir(working_dir)) == str(tmp_path)


def test_discovery_skips_setup_py_dir_with_init_py(tmp_path):
    (tmp_path / "pyproject.toml").touch()

    module_dir = tmp_path / "subdir1"
    module_dir.mkdir()

    (module_dir / "setup.py").touch()
    (module_dir / "__init__.py").touch()

    assert str(discover_project_dir(module_dir)) == str(tmp_path)

    # removing the __init__.py changes discovery
    (module_dir / "__init__.py").unlink()
    assert str(discover_project_dir(module_dir)) == str(module_dir)


def test_discovery_selects_setup_py_at_repo_root_as_final_guess(tmp_path):
    # "repo root"
    (tmp_path / "setup.py").touch()
    (tmp_path / ".git").mkdir()
    # but it contains an '__init__.py', so normally it would not be considered
    (tmp_path / "__init__.py").touch()

    subdir = tmp_path / "subdir1"
    subdir.mkdir()

    assert str(discover_project_dir(subdir)) == str(tmp_path)



import click

from mddj._cli.state import CommandState, common_args

from .version import self_version


@click.group("self")
@common_args
def self(*, state: CommandState) -> None:
    """Interact with mddj itself."""


self.add_command(self_version)



import re
from textwrap import dedent as d

import pytest


@pytest.mark.parametrize("lower_bound", (True, False))
def test_read_python_requires(chdir, tmp_path, run_line, lower_bound):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = "1.0.0"
            authors = [
              { name = "Foo", email = "foo@example.org" },
            ]
            requires-python = ">=3.11"
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        cmd = ["mddj", "read", "requires-python"]
        if lower_bound:
            cmd.append("--lower-bound")

        expect_result = r"^3\.11$" if lower_bound else r"^>=3\.11$"
        run_line(cmd, search_stdout=expect_result)


@pytest.mark.parametrize("quote_char", ('"', "'"))
def test_update_version_assignment(chdir, tmp_path, run_line, quote_char):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            f"""\
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = {quote_char}1.0.0{quote_char}
            authors = [
              {{ name = "Foo", email = "foo@example.org" }},
            ]
            """
        ),
        encoding="utf-8",
    )

    with chdir(tmp_path):
        run_line("mddj write version 2.3.1")

    assert pyproject.read_text() == d(
        f"""\
        [build-system]
        requires = ["setuptools"]
        build-backend = "setuptools.build_meta"

        [project]
        name = "foopkg"
        version = {quote_char}2.3.1{quote_char}
        authors = [
          {{ name = "Foo", email = "foo@example.org" }},
        ]
        """
    )


def test_read_version_from_pyproject(chdir, tmp_path, run_line):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = "8.0.7"
            authors = [
              { name = "Foo", email = "foo@example.org" },
            ]
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "foopkg.py").write_text("", encoding="utf-8")

    with chdir(tmp_path):
        run_line("mddj read version", search_stdout=r"^8\.0\.7$")


@pytest.mark.parametrize(
    ("version", "attr", "result"),
    (
        ("9.7.5", "major", "9"),
        ("9.7.5", "minor", "7"),
        ("9.7.5", "micro", "5"),
        ("9.7.5a3", "pre", "a3"),
        ("9.7.5b3", "pre", "b3"),
        ("9.7.5rc3", "pre", "rc3"),
        ("9.7.5post1", "post", "1"),
        ("9.7.5.post1", "post", "1"),
        ("9.7.5rc1", "release", "9.7.5"),
    ),
)
def test_read_version_attribute_from_pyproject(
    chdir, tmp_path, run_line, version, attr, result
):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            f"""\
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = "{version}"
            authors = [
              {{ name = "Foo", email = "foo@example.org" }},
            ]
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "foopkg.py").write_text("", encoding="utf-8")

    with chdir(tmp_path):
        run_line(
            f"mddj read version --attr {attr}",
            search_stdout="^" + re.escape(result) + "$",
        )


@pytest.mark.parametrize(
    ("version", "attr", "message"),
    (
        pytest.param("9.7.5", "pre", "'9.7.5' is not a prerelease", id="pre"),
        pytest.param("2.0-dev1", "post", "'2.0-dev1' is not a post-release", id="post"),
    ),
)
def test_read_version_attribute_from_pyproject_fails_due_to_type(
    chdir, tmp_path, run_line, version, attr, message
):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            f"""\
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = "{version}"
            authors = [
              {{ name = "Foo", email = "foo@example.org" }},
            ]
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "foopkg.py").write_text("", encoding="utf-8")

    with chdir(tmp_path):
        run_line(
            f"mddj read version --attr {attr}",
            search_stderr="^" + re.escape(message) + "$",
            assert_exit_code=1,
        )


@pytest.mark.parametrize(
    "bad_toml",
    (
        pytest.param(
            """\
            tool = 1
            """,
            id="tool-not-a-table",
        ),
        pytest.param(
            """\
            tool.mddj = 1
            """,
            id="tool.mddj-not-a-table",
        ),
    ),
)
def test_read_version_from_pyproject_ignores_malformed_tool_config(
    tmpdir, run_line, bad_toml
):
    pyproject = tmpdir.join("pyproject.toml")

    pyproject.write(
        d(
            f"""\
            {bad_toml}

            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = "8.0.7"
            authors = [
              {{ name = "Foo", email = "foo@example.org" }},
            ]
            """
        )
    )
    tmpdir.join("foopkg.py").write("")

    with tmpdir.as_cwd():
        run_line("mddj read version", search_stdout=r"^8\.0\.7$")


def test_read_dependencies(chdir, tmp_path, run_line):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"

            [project]
            name = "foopkg"
            version = "1.0.0"
            authors = [
              { name = "Foo", email = "foo@example.org" },
            ]
            dependencies = ["foo", "bar<2"]
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        result = run_line("mddj read dependencies")
        assert result.stdout == "foo\nbar<2\n"



import click

from mddj._cli.state import CommandState, common_args

from .version import write_version


@click.group("write")
@common_args
def write(*, state: CommandState) -> None:
    """Write metadata for the current project."""


write.add_command(write_version)



from textwrap import dedent as d

import pytest

from mddj._internal import _cached_toml
from mddj.api.reader import _read_pyproject_toml_value


@pytest.fixture
def document_cache():
    return _cached_toml.TomlDocumentCache()


def test_read_string_table_key(tmp_path, document_cache):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [project]
            version = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    read_val = _read_pyproject_toml_value(
        document_cache.load(pyproject), "project", "version"
    )
    assert read_val == "1.0.0"


def test_read_array_members(tmp_path, document_cache):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [mytable]
            items = [
                "foo",
                "bar",
            ]
            """
        ),
        encoding="utf-8",
    )

    read_val = _read_pyproject_toml_value(
        document_cache.load(pyproject), "mytable", "items", 0
    )
    assert read_val == "foo"


def test_read_bad_lookup_noncontainer(tmp_path, document_cache):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [mytable]
            foo = "bar"
            """
        ),
        encoding="utf-8",
    )

    with pytest.raises(LookupError, match="Terminated in a non-container type"):
        _read_pyproject_toml_value(
            document_cache.load(pyproject), "mytable", "foo", "bar"
        )


def test_read_bad_lookup_wrong_index_type(tmp_path, document_cache):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [mytable]
            foo = ["bar"]
            """
        ),
        encoding="utf-8",
    )

    with pytest.raises(LookupError, match="Incorrect index type"):
        _read_pyproject_toml_value(
            document_cache.load(pyproject), "mytable", "foo", "bar"
        )



import click

from mddj._cli.state import CommandState, common_args


@click.command("dependencies")
@common_args
def read_dependencies(*, state: CommandState) -> None:
    """Read the dependencies of the current project."""
    for d in state.dj.read.dependencies():
        click.echo(d)



from textwrap import dedent as d

import pytest


def test_read_python_requires(chdir, tmp_path, run_line, capfd):
    setupcfg = tmp_path / "setup.cfg"

    setupcfg.write_text(
        d(
            """\
            [metadata]
            name = foopkg
            version = 1.0.0

            author = Foo
            author_email = foo@example.org

            [options]
            python_requires = >=3.10
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "setup.py").write_text(
        "from setuptools import setup; setup()\n", encoding="utf-8"
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        result = run_line("mddj read requires-python", search_stdout=r"^>=3\.10$")
        assert len(result.stdout.splitlines()) == 1
    captured = capfd.readouterr()
    assert captured.out == ""


def test_read_python_requires_with_full_build_output_shows_all_data(
    chdir, tmp_path, run_line, capfd
):
    setupcfg = tmp_path / "setup.cfg"

    setupcfg.write_text(
        d(
            """\
            [metadata]
            name = foopkg
            version = 1.0.0

            author = Foo
            author_email = foo@example.org

            [options]
            python_requires = >=3.10
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "setup.py").write_text(
        "from setuptools import setup; setup()\n", encoding="utf-8"
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        result = run_line(
            "mddj read requires-python",
            search_stdout=r"^>=3\.10$",
            env={"MDDJ_CAPTURE_BUILD_OUTPUT": "false"},
        )
        assert len(result.stdout.splitlines()) == 1

    captured = capfd.readouterr()
    assert len(captured.out.splitlines()) > 1
    assert "running dist_info" in captured.out


def test_read_version_from_build(chdir, tmp_path, run_line):
    setupcfg = tmp_path / "setup.cfg"

    setupcfg.write_text(
        d(
            """\
            [metadata]
            name = foopkg
            version = 1.0.0

            author = Foo
            author_email = foo@example.org
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "setup.py").write_text(
        "from setuptools import setup; setup()\n", encoding="utf-8"
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        run_line("mddj read version", search_stdout=r"^1\.0\.0$")


def test_read_version_from_build_with_pyproject_present(chdir, tmp_path, run_line):
    setupcfg = tmp_path / "setup.cfg"
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(
        d(
            """\
            [project]
            name = "foopkg"
            dynamic = ["version"]
            """
        )
    )

    setupcfg.write_text(
        d(
            """\
            [metadata]
            name = foopkg
            version = 1.0.0

            author = Foo
            author_email = foo@example.org
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "setup.py").write_text(
        "from setuptools import setup; setup()\n", encoding="utf-8"
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        run_line("mddj read version", search_stdout=r"^1\.0\.0$")


@pytest.mark.parametrize("quote_char", ("", '"', "'"))
def test_update_version_assignment(chdir, tmp_path, run_line, quote_char):
    setupcfg = tmp_path / "setup.cfg"
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(
        d(
            """\
            [tool.mddj]
            write_version = "assign:setup.cfg:version"
            """
        ),
        encoding="utf-8",
    )

    setupcfg.write_text(
        d(
            f"""\
            [metadata]
            name = foopkg
            version = {quote_char}1.0.0{quote_char}
            author = Foo
            author_email = foo@example.org
            """
        ),
        encoding="utf-8",
    )

    with chdir(tmp_path):
        run_line("mddj write version 1.0.1")

    assert setupcfg.read_text() == d(
        f"""\
        [metadata]
        name = foopkg
        version = {quote_char}1.0.1{quote_char}
        author = Foo
        author_email = foo@example.org
        """
    )


def test_read_dependencies(chdir, tmp_path, run_line, capfd):
    setupcfg = tmp_path / "setup.cfg"

    setupcfg.write_text(
        d(
            """\
            [metadata]
            name = foopkg
            version = 1.0.0

            author = Foo
            author_email = foo@example.org

            [options]
            install_requires =
                foo
                bar<2
            """
        ),
        encoding="utf-8",
    )
    (tmp_path / "setup.py").write_text(
        "from setuptools import setup; setup()\n", encoding="utf-8"
    )
    (tmp_path / "foopkg.py").touch()

    with chdir(tmp_path):
        result = run_line("mddj read dependencies")
        assert result.stdout == "foo\nbar<2\n"
    captured = capfd.readouterr()
    assert captured.out == ""



"""
self-tests of mddj
"""

import os
import pathlib

import pytest

REPO_ROOT = pathlib.Path(__file__).parent.parent.parent


@pytest.fixture(autouse=True)
def _in_repo_root():
    old_cwd = os.getcwd()
    os.chdir(REPO_ROOT)
    try:
        yield
    finally:
        os.chdir(old_cwd)


def test_requires_python(run_line):
    run_line("mddj read requires-python", search_stdout=r"^>=3\.10$")



from __future__ import annotations

import functools
import pathlib

from .._internal import _cached_toml, _discovery
from .config import DJConfig, ReaderConfig, WriterConfig
from .reader import Reader
from .writer import Writer


class DJ:
    """
    An MDDJ DJ object is the top level access point for the library API.

    It provides programmatic access to the capabilities of MDDJ.

    Note that DJs and their components aggressively cache data so that it can be read
    many times quickly. In order to refresh state, instantiate a new DJ.
    """

    def __init__(self, config: DJConfig | None = None) -> None:
        self.config = config or DJConfig()
        self._document_cache = _cached_toml.TomlDocumentCache()

    @functools.cached_property
    def project_directory(self) -> pathlib.Path:
        """
        The directory where project metadata can be found.

        By default, this is automatically discovered on first access, but it can also be
        explicitly set via config.
        """
        if self.config.project_dir is not None:
            return self.config.project_dir

        return _discovery.discover_project_dir(self.config.discovery_start_dir)

    @functools.cached_property
    def pyproject_path(self) -> pathlib.Path:
        """
        The path to the pyproject.toml file which will be used to read and write
        metadata. The file may not actually exist, and this will not raise errors.
        """
        return self.project_directory / "pyproject.toml"

    @functools.cached_property
    def read(self) -> Reader:
        """A Reader configured via this DJ."""
        config = ReaderConfig(
            project_directory=self.project_directory,
            pyproject_path=self.pyproject_path,
            isolated_builds=self.config.isolated_builds,
            capture_build_output=self.config.capture_build_output,
        )
        return Reader(config, document_cache=self._document_cache)

    @functools.cached_property
    def write(self) -> Writer:
        """A Writer configured via this DJ."""
        config = WriterConfig.load_from_toml(
            project_directory=self.project_directory,
            pyproject_path=self.pyproject_path,
            document_cache=self._document_cache,
        )
        return Writer(config, document_cache=self._document_cache)



import click
from packaging.version import Version

from mddj._cli.state import CommandState, common_args


@click.command("list-versions")
@common_args
def tox_list_versions(*, state: CommandState) -> None:
    """Print all of the python versions tested under tox in version-sorted order."""
    versions = state.dj.read.tox.list_python_versions()
    for v in sorted(Version(v) for v in versions):
        click.echo(str(v))



from textwrap import dedent as d

import pytest

from mddj._internal import _writers


def test_write_table_key(tmp_path):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [project]
            version = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    write_result = _writers.write_toml_value(pyproject, "project.version", "1.1.0")
    assert write_result == "1.0.0"
    assert pyproject.read_text() == d(
        """\
        [project]
        version = "1.1.0"
        """
    )


def test_write_inline_table_key(tmp_path):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            project = {version = "1.0.0"}
            """
        ),
        encoding="utf-8",
    )

    write_result = _writers.write_toml_value(pyproject, "project.version", "1.1.0")
    assert write_result == "1.0.0"
    assert pyproject.read_text() == d(
        """\
        project = {version = "1.1.0"}
        """
    )


def test_write_at_top_level(tmp_path):
    config = tmp_path / "config.toml"

    config.write_text(
        d(
            """\
            version = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    write_result = _writers.write_toml_value(config, "version", "1.1.0")
    assert write_result == "1.0.0"
    assert config.read_text() == d(
        """\
        version = "1.1.0"
        """
    )


def test_write_array_element(tmp_path):
    config = tmp_path / "config.toml"

    config.write_text(
        d(
            """\
            flags = [
                "--verbose",
                "--okay",
            ]
            """
        ),
        encoding="utf-8",
    )

    write_result = _writers.write_toml_value(config, "flags.0", "--quiet")
    assert write_result == "--verbose"
    assert config.read_text() == d(
        """\
        flags = [
            "--quiet",
            "--okay",
        ]
        """
    )


def test_write_aot_table(tmp_path):
    config = tmp_path / "config.toml"

    config.write_text(
        d(
            """\
            [[projects]]
            version = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    write_result = _writers.write_toml_value(config, "projects.0.version", "1.1.0")
    assert write_result == "1.0.0"
    assert config.read_text() == d(
        """\
        [[projects]]
        version = "1.1.0"
        """
    )


def test_error_on_empty_path():
    with pytest.raises(ValueError, match="Cannot traverse an empty TOML path"):
        _writers.write_toml_value("foo.toml", "", "bar")


def test_top_level_key_must_be_str(tmp_path):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [project]
            version = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    with pytest.raises(
        ValueError, match="The first key of a TOML path must be a string"
    ):
        _writers.write_toml_value(pyproject, "1.bar", "baz")


def test_terminal_non_string_lookup_error(tmp_path):
    pyproject = tmp_path / "pyproject.toml"

    pyproject.write_text(
        d(
            """\
            [project]
            version = ["1.0.0"]
            """
        ),
        encoding="utf-8",
    )

    with pytest.raises(LookupError, match="TOML path terminated in a non-string value"):
        _writers.write_toml_value(pyproject, "project.version", "2.0")


def test_scalar_value_at_destination(tmp_path):
    config = tmp_path / "config.toml"

    config.write_text(
        d(
            """\
            [foo]
            bar = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    with pytest.raises(
        LookupError,
        match="Nontrivial TOML write path must terminate in a table or array",
    ):
        _writers.write_toml_value(config, "foo.bar.baz", "2.0")


def test_traversal_crosses_scalar_value(tmp_path):
    config = tmp_path / "config.toml"

    config.write_text(
        d(
            """\
            [foo]
            bar = "1.0.0"
            """
        ),
        encoding="utf-8",
    )

    with pytest.raises(
        LookupError, match=r"TOML path attempted to traverse scalar at \['foo', 'bar'\]"
    ):
        _writers.write_toml_value(config, "foo.bar.baz.quux", "2.0")



from textwrap import dedent as d


def test_read_min_version(chdir, tmp_path, run_line):
    toxini = tmp_path / "tox.ini"

    toxini.write_text(
        d(
            """\
            [tox]
            envlist = py{36,37,38,35,39,310}

            [testenv]
            commands = python -m pytest
            """
        ),
        encoding="utf-8",
    )

    with chdir(tmp_path):
        run_line("mddj read tox min-version", search_stdout="3.5")


def test_read_version_list(chdir, tmp_path, run_line):
    toxini = tmp_path / "tox.ini"

    toxini.write_text(
        d(
            """\
            [tox]
            envlist = py{36,37,38,35,39,310}

            [testenv]
            commands = python -m pytest
            """
        ),
        encoding="utf-8",
    )

    with chdir(tmp_path):
        result = run_line("mddj read tox list-versions")

    assert result.stdout == "3.5\n3.6\n3.7\n3.8\n3.9\n3.10\n"


def test_read_version_list_with_repeats_and_factors(chdir, tmp_path, run_line):
    toxini = tmp_path / "tox.ini"

    toxini.write_text(
        d(
            """\
            [tox]
            envlist =
                py{36,37,38,35,39,310}
                py{35,310}-toml

            [testenv]
            commands = python -m pytest
            """
        ),
        encoding="utf-8",
    )

    with chdir(tmp_path):
        result = run_line("mddj read tox list-versions")

    assert result.stdout == "3.5\n3.6\n3.7\n3.8\n3.9\n3.10\n"



import re
import shlex
import textwrap
from textwrap import dedent as d

import pytest
from click.testing import CliRunner

_PYTEST_VERBOSE = False


def pytest_configure(config):
    if config.getoption("verbose") > 0:
        global _PYTEST_VERBOSE
        _PYTEST_VERBOSE = True


@pytest.fixture
def cli_runner():
    return CliRunner()


@pytest.fixture
def run_line(cli_runner):
    def func(
        line,
        assert_exit_code=0,
        stdin=None,
        search_stdout=None,
        search_stderr=None,
        env=None,
    ):
        from mddj._cli import main

        # split line into args and confirm line starts with "mddj"
        args = shlex.split(line) if isinstance(line, str) else line
        assert args[0] == "mddj"

        # run the line. main is the "mddj" part of the line
        # if we are expecting success (0), don't catch any exceptions.
        result = cli_runner.invoke(
            main,
            args[1:],
            input=stdin,
            catch_exceptions=bool(assert_exit_code),
            env=env,
        )
        if result.exit_code != assert_exit_code:
            message = d(
                f"""\
                CliTest run_line exit_code assertion failed!
                Line:
                  {line}
                exited with {result.exit_code} when expecting {assert_exit_code}
                """
            )
            if _PYTEST_VERBOSE:
                message += (
                    "\n\nstdout:\n"
                    + textwrap.indent(result.stdout, "  ")
                    + "\nstderr:"
                    + textwrap.indent(result.stderr, "  ")
                )

            pytest.fail(message)
        if search_stdout is not None:
            _assert_matches(result.stdout, "stdout", search_stdout)
        if search_stderr is not None:
            _assert_matches(result.stderr, "stderr", search_stderr)
        return result

    return func


def _assert_matches(text, text_name, search):
    __tracebackhide__ = True

    if isinstance(search, (str, re.Pattern)):
        search = [search]
    elif not isinstance(search, list):
        raise NotImplementedError(
            "search_{stdout,stderr} got unexpected arg type: {type(search)}"
        )

    compiled_searches = [
        s if isinstance(s, re.Pattern) else re.compile(s, re.MULTILINE) for s in search
    ]
    for pattern in compiled_searches:
        if pattern.search(text) is None:
            if _PYTEST_VERBOSE:
                pytest.fail(
                    f"Pattern('{pattern.pattern}') not found in {text_name}.\n"
                    f"Full text:\n\n{text}"
                )
            else:
                pytest.fail(
                    f"Pattern('{pattern.pattern}') not found in {text_name}. "
                    "Use 'pytest -v' to see full output."
                )



from __future__ import annotations

import pathlib
import typing as t


def discover_project_dir(start_dir: pathlib.Path) -> pathlib.Path:
    for directory, dir_contents in _vcs_bounded_ancestors_with_contents(start_dir):
        # Rule 1: the first ancestor dir with a `pyproject.toml`
        if "pyproject.toml" in dir_contents:
            return directory

        # Rule 2: a `setup.cfg`
        if "setup.cfg" in dir_contents:
            return directory

        # Rule 3: `setup.py` in a directory without an `__init__.py`
        if "setup.py" in dir_contents and "__init__.py" not in dir_contents:
            return directory

        # Rule 4: a `tox.ini` file is present, meaning we can read tox data
        #         (maybe there is no python package but there is tox config)
        if "tox.ini" in dir_contents:
            return directory

    # Rule 5: the root has a `setup.py`; that will be our final guess
    if "setup.py" in dir_contents:
        return directory

    raise LookupError(
        "mddj could not find the project root. "
        "Ensure you are running from a subdirectory of a Python package source dir."
    )


def _vcs_bounded_ancestors_with_contents(
    start_dir: pathlib.Path,
) -> t.Iterator[tuple[pathlib.Path, set[str]]]:
    for d in _ancestors(start_dir):
        dir_contents = {p.name for p in d.iterdir()}
        if dir_contents.intersection((".git", ".hg", ".svn")):
            yield d, dir_contents
            return
        yield d, dir_contents


def _ancestors(start_dir: pathlib.Path) -> t.Iterator[pathlib.Path]:
    yield start_dir
    yield from start_dir.parents



from __future__ import annotations

import pathlib
import re

import tomlkit

from ._toml_path import parse_toml_path
from ._types import TomlValue, is_toml_array, is_toml_mapping, is_toml_table


def write_simple_assignment(
    path: str | pathlib.Path, key: str, value: str
) -> str | None:
    if isinstance(path, str):
        path = pathlib.Path(path)

    line_pattern = re.compile(rf"^\s*{re.escape(key)}\s*=\s*(.*)$")
    with path.open("r", encoding="utf-8") as fp:
        content = fp.readlines()

    old_value = None
    with path.open("w", encoding="utf-8") as fp:
        for line in content:
            if m := line_pattern.match(line):
                old_value = m.group(1)
                if old_value.startswith('"') and old_value.endswith('"'):
                    new_value = f'"{value}"'
                elif old_value.startswith("'") and old_value.endswith("'"):
                    new_value = f"'{value}'"
                else:
                    new_value = value

                fp.write(f"{key} = {new_value}\n")
            else:
                fp.write(line)

    if old_value is None:
        return None
    else:
        return old_value.strip("'\"")


def write_toml_value(
    path: str | pathlib.Path,
    target: str,
    value: str,
    *,
    loaded_document: tomlkit.TOMLDocument | None = None,
) -> str:
    """
    Write a value to a given target (as a string) in a TOML file.

    The value must be a string and the TOML traversal must satisfy
    """
    if isinstance(path, str):
        path = pathlib.Path(path)

    target_path = parse_toml_path(target)
    if not target_path:
        msg = "Cannot traverse an empty TOML path."
        raise ValueError(msg)

    if loaded_document is not None:
        doc = loaded_document
    else:
        with path.open("r", encoding="utf-8") as read_file_descriptor:
            doc = tomlkit.load(read_file_descriptor)

    write_container, key = _traverse_toml_path_to_write_point(doc, target_path)

    if is_toml_mapping(write_container):
        if isinstance(key, str):
            old_value = write_container[key]
        else:
            msg = "TOML traversal found subtable, but final key in path was an int."
            raise LookupError(msg)
    elif is_toml_array(write_container):
        if isinstance(key, int):
            old_value = write_container[key]
        else:
            msg = "TOML traversal found subtable, but final key in path was a string."
            raise LookupError(msg)
    else:
        msg = (
            "Nontrivial TOML write path must terminate in a table or array, "
            "got scalar."
        )
        raise LookupError(msg)

    if old_value.value == value:
        return value
    elif isinstance(old_value, tomlkit.items.String):
        # check string quoting style
        # update when the change for https://github.com/python-poetry/tomlkit/issues/443
        # is released
        if old_value._t.is_literal():
            write_value = tomlkit.string(value, literal=True)
        else:
            write_value = tomlkit.string(value)

        # mypy flags this, but we've validated the key type against the container type
        # above -- the narrowed type info is simply not retained
        write_container[key] = write_value  # type: ignore[index]
        with path.open("w", encoding="utf-8") as write_file_descriptor:
            tomlkit.dump(doc, write_file_descriptor)
        return old_value.value
    else:
        msg = "TOML path terminated in a non-string value."
        raise LookupError(msg)


def _traverse_toml_path_to_write_point(
    doc: tomlkit.TOMLDocument, path: list[str | int]
) -> tuple[TomlValue, str | int]:
    """
    Traverse a path in a doc terminating in a table or array, returning the final key
    used to write into that subdoc.

    :raises ValueError: if the path does not match the document shape
    """
    if not path:
        msg = "Cannot traverse an empty TOML path."
        raise ValueError(msg)

    first_key, remainder = path[0], path[1:]
    if not isinstance(first_key, str):
        msg = "The first key of a TOML path must be a string."
        raise ValueError(msg)

    if not remainder:
        return (doc, first_key)

    remainder, final_key = remainder[:-1], remainder[-1]

    element: TomlValue = doc[first_key]
    debug_path: list[int | str] = [first_key]
    for key in remainder:
        if not (is_toml_array(element) or is_toml_table(element)):
            msg = f"TOML path attempted to traverse scalar at {debug_path}"
            raise LookupError(msg)

        if isinstance(key, int) and is_toml_array(element):  # slyp: disable=W200
            element = element[key]
        elif isinstance(key, str) and is_toml_table(element):
            element = element[key]
        else:
            msg = (
                "Error traversing TOML document. "
                f"Encountered improper type at {debug_path}"
            )
            raise LookupError(msg)
        debug_path.append(key)

    return element, final_key



from __future__ import annotations

import collections.abc
import functools
import pathlib

import tomlkit

from .._internal import _cached_toml, _types
from .config import ReaderConfig
from .tox_reader import ToxReader

try:
    import importlib_metadata as _importlib_metadata
except ImportError:
    import importlib.metadata as _importlib_metadata


class Reader:
    """
    A Reader is an interface for data reading capabilities.

    Typically, users should simply create a DJ and then access the reader built by it,
    as in:

    .. code-block:: pycon

        >>> from mddj.api import DJ
        >>> dj = DJ()
        >>> dj.read.version()
        '0.1.0'

    :ivar tox: a :class:`ToxReader` provided by this reader
    """

    def __init__(
        self,
        config: ReaderConfig,
        document_cache: _cached_toml.TomlDocumentCache | None = None,
    ) -> None:
        self.config = config
        self._document_cache = document_cache or _cached_toml.TomlDocumentCache()

        self.tox = ToxReader()

    @functools.cached_property
    def _wheel_metadata(self) -> _importlib_metadata.PackageMetadata:
        return _get_wheel_metadata(
            self.config.project_directory,
            isolated=self.config.isolated_builds,
            quiet=self.config.capture_build_output,
        )

    @functools.cached_property
    def _pyproject_toml_document(self) -> tomlkit.TOMLDocument:
        return self._document_cache.load(self.config.pyproject_path)

    @functools.cached_property
    def _pyproject_dynamic(self) -> tuple[str, ...]:
        try:
            value = _read_pyproject_toml_value(
                self._pyproject_toml_document, "project", "dynamic"
            )
        except (FileNotFoundError, LookupError):
            return ()
        if _types.is_toml_array(value) and all(isinstance(x, str) for x in value):
            return tuple(value)
        raise LookupError("project.dynamic must be an array of strings when present")

    def _read_static(self, key: str) -> object | None:
        if key in self._pyproject_dynamic:
            return None

        try:
            return _read_pyproject_toml_value(
                self._pyproject_toml_document, "project", key
            )
        except (FileNotFoundError, LookupError):
            return None

    def version(self) -> str:
        """Get the version of the project."""
        value = self._read_static("version")
        if value is not None:
            return str(value)
        return str(self._wheel_metadata.get("Version"))

    @functools.cached_property
    def _requires_python(self) -> str:
        value = self._read_static("requires-python")
        if value is not None:
            return str(value)

        value = self._wheel_metadata.get("Requires-Python")
        if value is None:
            raise LookupError("No Requires-Python data found")
        return str(value)

    def requires_python(self, *, lower_bound: bool = False) -> str:
        """
        Get the Requires-Python bound for the project.

        :param lower_bound: When true, only get the lower bound of a `>=` expression.
            Only `>=` is supported. Any other comparator will produce a ValueError if
            `lower_bound` is set.
        """
        if lower_bound:
            return _requires_python_lower_bound(self._requires_python)
        return self._requires_python

    @functools.cached_property
    def _dependencies(self) -> tuple[str, ...]:
        value: object | None = self._read_static("dependencies")

        if _types.is_toml_array(value):
            return tuple(str(d) for d in value)

        return tuple(str(d) for d in self._wheel_metadata.get_all("Requires-Dist"))

    def dependencies(self) -> tuple[str, ...]:
        """Get the dependencies for the project."""
        return self._dependencies


def _requires_python_lower_bound(req: str) -> str:
    reqs = req.split(",")
    lower_bounds = []
    for r in reqs:
        if r.startswith(">="):
            lower_bounds.append(r[2:])
        elif r.startswith(">"):
            raise ValueError(
                "Found a > requirement, rejecting as invalid (use '>=' instead)"
            )
    if len(lower_bounds) > 1:
        raise ValueError("Found multiple lower bounds, cannot choose one")
    elif len(lower_bounds) == 0:
        raise ValueError("Found no lower bounds")
    else:
        return lower_bounds[0]


def _get_wheel_metadata(
    source_dir: pathlib.Path, isolated: bool = True, quiet: bool = True
) -> _importlib_metadata.PackageMetadata:
    """
    Get metadata for wheel, either using the PEP 517 hook or by actually
    doing a wheel build and examining the result.
    """
    import build.util
    import pyproject_hooks

    runner = pyproject_hooks.quiet_subprocess_runner
    if not quiet:
        runner = pyproject_hooks.default_subprocess_runner
    return build.util.project_wheel_metadata(
        source_dir, isolated=isolated, runner=runner
    )


def _read_pyproject_toml_value(
    pyproject_data: tomlkit.TOMLDocument, *path: str | int
) -> object:
    """
    Read an arbitrary value from 'pyproject.toml'
    """
    # traverse the TOML data structure
    cursor: _types.TomlValue = pyproject_data
    for subkey in path:
        # pedantically enumerate the branches for static type checking to
        # easily see the association between key and container types
        if isinstance(subkey, str) and _types.is_toml_mapping(  # slyp: disable=W200
            cursor
        ):
            cursor = cursor[subkey]
        elif isinstance(subkey, int) and _types.is_toml_array(cursor):
            cursor = cursor[subkey]
        else:
            message = f"Could not lookup '{path}' in pyproject.toml."
            # str is a container and a scalar...
            if isinstance(cursor, str) or not isinstance(
                cursor, collections.abc.Container
            ):
                message = f"{message} Terminated in a non-container type."
            else:
                message = f"{message} Incorrect index type."
            raise LookupError(message)

    return cursor



from importlib.metadata import version

import click

from mddj._cli.state import CommandState, common_args


@click.command("version")
@common_args
def self_version(*, state: CommandState) -> None:
    """Get the version of mddj."""
    click.echo("mddj version: " + version("mddj"))



import click

from mddj._cli.state import CommandState, common_args


@click.command("version")
@click.argument("NEW_VERSION")
@common_args
def write_version(*, new_version: str, state: CommandState) -> None:
    """Write the 'version' of the current project, based on config."""
    try:
        result = state.dj.write.version(new_version)
    except LookupError as e:
        click.echo(str(e))
        click.get_current_context().exit(1)
    click.echo(
        f"""\
Version was updated. write_version='{state.dj.write.config.write_version}'

old value: {result}
new value: {new_version}
"""
    )



from __future__ import annotations

import dataclasses
import functools
import os
import pathlib
import sys
import typing as t

from .._internal import _cached_toml

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

_WRITE_VERSION_MODES = ("assign", "toml")


def _bool_env_var_default_factory(varname: str, default: bool) -> t.Callable[[], bool]:
    def factory() -> bool:
        if (var := os.environ.get(varname)) is not None:
            return var.lower() not in ("0", "false")
        return default

    return factory


@dataclasses.dataclass
class DJConfig:
    """
    Configuration for a DJ object.

    Some values support having their defaults set via environment variables:

    - ``isolated_builds``: ``MDDJ_ISOLATED_BUILDS``
    - ``capture_build_output``: ``MDDJ_CAPTURE_BUILD_OUTPUT``
    """

    #: The starting directory for project dir discovery. Defaults to cwd.
    discovery_start_dir: pathlib.Path = dataclasses.field(
        default_factory=pathlib.Path.cwd
    )
    #: The project directory. By default, it is discovered from the start directory.
    project_dir: pathlib.Path | None = None
    #: Whether or not to use isolated builds when getting metadata from build backends.
    #: Defaults to True.
    isolated_builds: bool = dataclasses.field(
        default_factory=_bool_env_var_default_factory("MDDJ_ISOLATED_BUILDS", True)
    )
    #: Whether or not to use "quiet mode" for builds when they are invoked.
    #: Defaults to True.
    capture_build_output: bool = dataclasses.field(
        default_factory=_bool_env_var_default_factory("MDDJ_CAPTURE_BUILD_OUTPUT", True)
    )


@dataclasses.dataclass
class ReaderConfig:
    """
    Configuration for a metadata reader.
    """

    project_directory: pathlib.Path
    pyproject_path: pathlib.Path
    isolated_builds: bool
    capture_build_output: bool


@dataclasses.dataclass
class WriterConfig:
    """
    Configuration for a metadata writer.
    """

    project_directory: pathlib.Path
    write_version: str = "toml: pyproject.toml: project.version"

    @classmethod
    def load_from_toml(
        cls,
        *,
        project_directory: pathlib.Path,
        pyproject_path: pathlib.Path,
        document_cache: _cached_toml.TomlDocumentCache,
    ) -> Self:
        import tomlkit

        if not pyproject_path.exists():
            return cls(project_directory=project_directory)

        data = document_cache.load(pyproject_path)

        try:
            tool_table = data["tool"]
            if not isinstance(tool_table, tomlkit.items.Table):
                raise KeyError("'tool' was not a table")
            mddj_data = tool_table["mddj"]
            if not isinstance(mddj_data, tomlkit.items.Table):
                raise KeyError("'tool.mddj' was not a table")

            write_version = mddj_data["write_version"]
            if not isinstance(write_version, str):
                raise KeyError("'tool.mddj.write_version' must be a str")
        except KeyError:
            return cls(project_directory=project_directory)

        return cls(project_directory=project_directory, write_version=write_version)

    @functools.cached_property
    def write_version_settings(self) -> WriteVersionSettings:
        write_version_mode, _, write_version_value = self.write_version.partition(":")

        if not write_version_value:
            raise ValueError("write_version must be of the form 'mode:value'.")

        if write_version_mode not in _WRITE_VERSION_MODES:
            raise ValueError(
                f"{write_version_mode} is not a valid mode for write_version. "
                f"Please choose one of {_WRITE_VERSION_MODES}"
            )

        if write_version_mode == "toml":
            return WriteVersionTomlSettings.parse(write_version_value)
        elif write_version_mode == "assign":
            return WriteVersionAssignSettings.parse(write_version_value)
        else:
            raise NotImplementedError(
                "Internal error. "
                f"Unimplemented write_version_mode: {write_version_mode}"
            )


WriteVersionSettings: t.TypeAlias = t.Union[
    "WriteVersionTomlSettings", "WriteVersionAssignSettings"
]


@dataclasses.dataclass
class WriteVersionTomlSettings:
    file_path: pathlib.Path
    toml_path: str

    @classmethod
    def parse(cls, config_str: str) -> Self:
        if ":" in config_str:
            file_path, _, toml_path = config_str.partition(":")
            file_path = file_path.strip()
            toml_path = toml_path.strip()
        else:
            file_path = "pyproject.toml"
            toml_path = config_str.strip()

        return cls(file_path=pathlib.Path(file_path), toml_path=toml_path)


@dataclasses.dataclass
class WriteVersionAssignSettings:
    file_path: pathlib.Path
    key: str

    @classmethod
    def parse(cls, config_str: str) -> Self:
        if ":" in config_str:
            file_path, _, key = config_str.partition(":")
            file_path = file_path.strip()
            key = key.strip()
        else:
            file_path = "pyproject.toml"
            key = config_str.strip()

        return cls(file_path=pathlib.Path(file_path), key=key)



import click

from .read import read
from .self import self
from .state import CommandState, common_args
from .write import write


@click.group("mddj")
@common_args
def main(*, state: CommandState) -> None:
    """MetaData DJ"""


main.add_command(read)
main.add_command(write)
main.add_command(self)



from __future__ import annotations

import functools
import typing as t

import click

from ..api import DJ

F = t.TypeVar("F", bound=t.Callable[..., t.Any])


class CommandState:
    def __init__(self) -> None:
        self.dj = DJ()


def common_args(cmd: F) -> F:
    @functools.wraps(cmd)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        state = click.get_current_context().ensure_object(CommandState)
        return cmd(*args, state=state, **kwargs)

    wrapper = click.help_option("-h", "--help")(wrapper)
    return t.cast(F, wrapper)
